import pytest
from django.conf import settings
from django.contrib.auth.models import User
from django.test import Client
from django.urls import resolve, reverse

from tests.models import MyModel


@pytest.mark.django_db
class TestIntegration:
    def test_app_installed(self):
        assert "changeform_actions" in settings.INSTALLED_APPS

    def test_urls_included(self):
        match = resolve(reverse("changeform_actions:run_admin_changeform_action"))
        assert match.app_name == "changeform_actions"
        assert match.url_name == "run_admin_changeform_action"

    def test_admin_mixin_importable(self):
        from changeform_actions import ChangeFormActionsMixin

        assert ChangeFormActionsMixin is not None

    def test_changeform_has_actions_dropdown(self):
        # Setup
        User.objects.create_superuser("admin", "admin@example.com", "password")
        client = Client()
        client.login(username="admin", password="password")

        instance = MyModel.objects.create(name="Test")

        # Test: Get changeform page and check that the dropdown is present
        changeform_url: str = reverse("admin:tests_mymodel_change", args=[instance.pk])
        response = client.get(changeform_url)
        assert response.status_code == 200
        assert b'<select name="action"' in response.content


@pytest.mark.django_db
class TestRunAction:
    def test_run_action_needs_staff_user_login(self):
        # Setup
        User.objects.create_user("user", "user@example.com", "password")
        client = Client()
        client.login(username="user", password="password")

        instance = MyModel.objects.create(name="Test")

        # Test: Redirected to admin login instead
        run_action_url: str = reverse("changeform_actions:run_admin_changeform_action")
        response = client.post(
            run_action_url,
            data={
                "app_label": "tests",
                "model_name": "mymodel",
                "pk": str(instance.pk),
                "action": "add_copy_suffix",
            },
            HTTP_REFERER=reverse("admin:tests_mymodel_change", args=[instance.pk]),
        )
        assert response.status_code == 302
        assert response.url.startswith(reverse("admin:login"))

    def test_run_action(self):
        # Setup
        User.objects.create_superuser("admin", "admin@example.com", "password")
        client = Client()
        client.login(username="admin", password="password")

        instance = MyModel.objects.create(name="Test")

        # Test: Run action renames the model instance
        run_action_url: str = reverse("changeform_actions:run_admin_changeform_action")
        response = client.post(
            run_action_url,
            data={
                "app_label": "tests",
                "model_name": "mymodel",
                "pk": str(instance.pk),
                "action": "add_copy_suffix",
            },
            HTTP_REFERER=reverse("admin:tests_mymodel_change", args=[instance.pk]),
            follow=True,
        )
        assert response.status_code == 200

        instance.refresh_from_db()
        assert instance.name == "Test - copy"



from django.urls import path

from . import views

# Register namespace for all URLs in this app
app_name = "changeform_actions"

urlpatterns = [
    path(
        "django-admin-changeform-actions/run",
        views.run_action_for_model_instance,
        name="run_admin_changeform_action",
    ),
]



from django.apps import apps
from django.contrib import admin
from django.contrib.admin.views.decorators import staff_member_required
from django.db.models import QuerySet
from django.http import HttpResponseRedirect
from django.utils.translation import gettext_lazy as _
from django.views.decorators.http import require_http_methods


@staff_member_required
@require_http_methods(("POST",))
def run_action_for_model_instance(request):
    """Run specified custom admin action on a model instance.

    This acts as a custom handler for our custom admin actions
    dropdown on a ModelAdmin's changeform page. The idea is to have the
    same outcome as selecting the model instance on the changelist page
    and executing the action there, just more conveniently from its
    changeform aka detail page.

    After running the action we redirect back to where we came from.

    For the corresponding ModelAdmin code, see:

        admin.ChangeFormActionsMixin

    POST Args:

        app_label: str = The app label of the model.

        model_name: str = The model name.

        pk: int = The primary key of the model instance.

        action: str = The action name to run.

    Returns:

        HttpResponseRedirect = Redirect back to the changeform page.

    """
    referer_url: str = request.META["HTTP_REFERER"]

    app_label: str = request.POST["app_label"]
    model_name: str = request.POST["model_name"]
    pk: int = int(request.POST["pk"])
    action_name: str = request.POST.get("action", "")

    if not action_name:
        # Probably submitted the default empty option, ignore.
        return HttpResponseRedirect(referer_url)

    model_cls: type = apps.get_model(app_label, model_name)
    queryset: QuerySet = model_cls.objects.filter(pk=pk)

    model_admin: object = admin.site._registry[model_cls]
    for action, name, label in model_admin._get_base_actions():
        if name == action_name:
            action(model_admin, request, queryset)
            model_admin.message_user(request, _("Action:") + f" {label}")

    return HttpResponseRedirect(referer_url)



from django.contrib import admin

from changeform_actions import ChangeFormActionsMixin
from tests.models import MyModel


class MyModelAdmin(ChangeFormActionsMixin, admin.ModelAdmin):
    actions = ["add_copy_suffix"]

    def add_copy_suffix(self, request, queryset):
        for instance in queryset:
            instance.name += " - copy"
            instance.save()


admin.site.register(MyModel, MyModelAdmin)



from django.template import Library

register = Library()


@register.simple_tag(takes_context=True)
def get_changeform_actions_dropdown(context, model_admin) -> str:
    """Return optional <select> HTML if ModelAdmin has been configured accordingly."""
    request = context["request"]
    try:
        return model_admin.get_changeform_actions_dropdown(request, context)
    except AttributeError:
        return ""



from django import forms
from django.utils.translation import gettext_lazy as _


class ActionForm(forms.Form):
    """Simplified version of django.contrib.admin.helpers.ActionForm."""

    action = forms.ChoiceField(label=_("Action:"))



from typing import Optional

from django.template.context import ContextDict
from django.template.loader import render_to_string
from django.urls import reverse

from .forms import ActionForm

DEFAULT_DELETE_ACTION_NAME: str = "delete_selected"


class ChangeFormActionsMixin:
    """Add support to run custom admin actions on the changeform page.

    This is handy as it allows to run custom actions on a single object
    instead of having to go back to the changelist page and select the
    desired one object there.

    Django's actions are not really made to be used outside of the
    changelist page, which is why this feature here is somewhat of a
    hack. It consists of the following:

    This mixin for a ModelAdmin to get the HTML for a dropdown:

      admin:ChangeFormActionsMixin

    A template tag to inject that dropdown into the changeform page:

      templatetags:get_changeform_actions_dropdown()

    A view to handle the dropdown submission and run the action:

      views:run_action_for_model_instance()

    We are bypassing the builtin actions handling, which expects a POST
    request against the ModelAdmin's changelist view including a
    selection of objects to act upon and an index for the selected
    action. Instead we pass any needed information as POST to our view,
    which then figures out which ModelAdmin instance to fetch from the
    admin.site registry and which action method to run on it against the
    current object/instance.

    """

    skip_default_delete_action: bool = True

    def get_form_action_url(self) -> str:
        """Override this to customize the form action URL.

        Please be aware you will then have to manually register the
        corresponding run_action_for_model_instance() view to that URL,
        instead of just including the provided urls.py.

        """
        return reverse("changeform_actions:run_admin_changeform_action")

    def get_changeform_actions_dropdown(self, request, context) -> str:
        """Return HTML to run custom admin actions on the current object."""

        opts: dict = self.model._meta
        app_label: str = opts.app_label
        model_name: str = opts.model_name

        object_id: Optional[int] = None
        for item in context:
            if not isinstance(item, ContextDict):
                continue
            if "object_id" in item:
                object_id = item["object_id"]
                break
        if not object_id:
            return ""

        action_form = ActionForm(auto_id=None)
        action_form.fields["action"].choices = [
            (name, label)
            for (name, label) in self.get_action_choices(request)
            if self.skip_default_delete_action and name != DEFAULT_DELETE_ACTION_NAME
        ]

        return render_to_string(
            "admin/actions_for_changeform.html",
            request=request,
            context={
                "action_form_url": self.get_form_action_url(),
                "action_form": action_form,
                "app_label": app_label,
                "model_name": model_name,
                "object_id": object_id,
            },
        )
